import os

from collections import deque
from typing import Callable

from clipped.formatting import Printer
from clipped.utils.json import orjson_loads
from clipped.utils.tz import local_datetime

from polyaxon import settings
from polyaxon._containers.names import MAIN_CONTAINER_NAMES
from traceml.logging.schemas import V1Log, V1Logs


def get_logs_streamer(
    show_timestamp: bool = True, all_containers: bool = False, all_info: bool = False
) -> Callable:
    colors = deque(Printer.COLORS)
    job_to_color = {}
    if all_info:
        all_containers = True

    def handle_log_line(log: V1Log):
        log_dict = log.to_dict()
        log_line = ""
        if log.timestamp and show_timestamp:
            date_value = local_datetime(
                log_dict.get("timestamp"), tz=settings.CLIENT_CONFIG.timezone
            )
            log_line = Printer.add_log_color(date_value, "white") + " | "

        def get_container_info():
            if container_info in job_to_color:
                color = job_to_color[container_info]
            else:
                color = colors[0]
                colors.rotate(-1)
                job_to_color[container_info] = color
            return Printer.add_log_color(container_info, color) + " | "

        if not all_containers and log.container not in MAIN_CONTAINER_NAMES:
            return log_line

        if all_info:
            container_info = ""
            if log.node:
                log_line += Printer.add_log_color(log_dict.get("node"), "white") + " | "
            if log.pod:
                log_line += Printer.add_log_color(log_dict.get("pod"), "white") + " | "
            if log.container:
                container_info = log_dict.get("container")

            log_line += get_container_info()

        log_line += log_dict.get("value")
        Printer.log(log_line, nl=True)

    def handle_log_lines(logs: V1Logs):
        for log in logs.logs:
            if log:
                handle_log_line(log=log)

    return handle_log_lines


def load_logs_from_path(
    logs_path: str,
    hide_time: bool = False,
    all_containers: bool = True,
    all_info: bool = True,
):
    for file_logs in sorted(os.listdir(logs_path)):
        with open(os.path.join(logs_path, file_logs)) as f:
            logs_data = orjson_loads(f.read()).get("logs", [])
            logs_stream = V1Logs(logs=logs_data)
            get_logs_streamer(
                show_timestamp=not hide_time,
                all_containers=all_containers,
                all_info=all_info,
            )(logs_stream)



import datetime

from typing import List, Optional, Text

from clipped.compact.pydantic import StrictStr
from clipped.utils.dates import parse_datetime
from clipped.utils.json import orjson_dumps, orjson_loads
from clipped.utils.strings import validate_file_or_buffer
from clipped.utils.tz import now

from polyaxon._schemas.base import BaseSchemaModel
from traceml.logging.parser import (
    DATETIME_REGEX,
    ISO_DATETIME_REGEX,
    timestamp_search_regex,
)


class V1Log(BaseSchemaModel):
    _SEPARATOR = "|"
    _IDENTIFIER = "log"

    timestamp: Optional[datetime.datetime]
    node: Optional[StrictStr]
    pod: Optional[StrictStr]
    container: Optional[StrictStr]
    value: Optional[StrictStr]

    @classmethod
    def process_log_line(
        cls,
        value: Text,
        node: Optional[str],
        pod: Optional[str],
        container: Optional[str],
        timestamp=None,
    ) -> Optional["V1Log"]:
        if not value:
            return None

        if not isinstance(value, str):
            value = value.decode("utf-8")

        value = value.strip()

        if not timestamp:
            value, timestamp = timestamp_search_regex(ISO_DATETIME_REGEX, value)
            if not timestamp:
                value, timestamp = timestamp_search_regex(DATETIME_REGEX, value)
        if isinstance(timestamp, str):
            try:
                timestamp = parse_datetime(timestamp)
            except Exception as e:
                raise ValueError("Received an invalid timestamp") from e

        return cls.construct(
            timestamp=timestamp if timestamp else now(tzinfo=True),
            node=node,
            pod=pod,
            container=container,
            value=value,
        )

    def to_csv(self) -> str:
        values = [
            str(self.timestamp) if self.timestamp is not None else "",
            str(self.node) if self.node is not None else "",
            str(self.pod) if self.pod is not None else "",
            str(self.container) if self.container is not None else "",
            orjson_dumps({"_": self.value}) if self.value is not None else "",
        ]

        return self._SEPARATOR.join(values)


class V1Logs(BaseSchemaModel):
    _CHUNK_SIZE = 2000
    _IDENTIFIER = "logs"

    logs: Optional[List[V1Log]]
    last_time: Optional[datetime.datetime]
    last_file: Optional[StrictStr]
    files: Optional[List[StrictStr]]

    @classmethod
    def get_csv_header(cls) -> str:
        return V1Log._SEPARATOR.join(V1Log.__fields__.keys())

    def to_csv(self):
        _logs = ["\n{}".format(e.to_csv()) for e in self.logs if e.value]
        return "".join(_logs)

    @classmethod
    def should_chunk(cls, logs: List[V1Log]):
        return len(logs) >= cls._CHUNK_SIZE

    @classmethod
    def chunk_logs(cls, logs: List[V1Log]):
        total_size = len(logs)
        for i in range(0, total_size, cls._CHUNK_SIZE):
            yield cls(logs=logs[i : i + cls._CHUNK_SIZE])

    @classmethod
    def read_csv(cls, data: str, parse_dates: bool = True) -> "V1Logs":
        import numpy as np
        import pandas as pd

        csv = validate_file_or_buffer(data)
        if parse_dates:
            df = pd.read_csv(
                csv,
                sep=V1Log._SEPARATOR,
                parse_dates=["timestamp"],
                error_bad_lines=False,
                engine="pyarrow",
            )
        else:
            df = pd.read_csv(
                csv,
                sep=V1Log._SEPARATOR,
                engine="pyarrow",
            )

        return cls.construct(
            logs=[
                V1Log.construct(
                    timestamp=i.get("timestamp"),
                    node=i.get("node"),
                    pod=i.get("pod"),
                    container=i.get("container"),
                    value=orjson_loads(i.get("value")).get("_"),
                )
                for i in df.replace({np.nan: None}).to_dict(orient="records")
            ]
        )



from traceml.exceptions import TracemlException
from traceml.run import Run

try:
    from ignite.contrib.handlers.polyaxon_logger import PolyaxonLogger
except ImportError:
    raise TracemlException("ignite is required to use the tracking Logger")


class Logger(PolyaxonLogger):
    def __init__(self, *args, **kwargs):
        self.experiment = kwargs.get("run")
        if not self.experiment:
            self.experiment = Run(*args, **kwargs)



from typing import Optional

from clipped.utils.enums import get_enum_value


def get_resource_path(
    run_path: str, kind: Optional[str] = None, name: Optional[str] = None
) -> str:
    _path = "{}/resources".format(run_path)
    if kind:
        _path = "{}/{}".format(_path, get_enum_value(kind))
    if name:
        _path = "{}/{}.plx".format(_path, name)

    return _path


def get_event_path(
    run_path: str, kind: Optional[str] = None, name: Optional[str] = None
) -> str:
    _path = "{}/events".format(run_path)
    if kind:
        _path = "{}/{}".format(_path, get_enum_value(kind))
    if name:
        _path = "{}/{}.plx".format(_path, name)

    return _path


def get_event_assets_path(run_path: str, kind: Optional[str] = None) -> str:
    _path = "{}/assets".format(run_path)
    if kind:
        _path = "{}/{}".format(_path, get_enum_value(kind))
    return _path


def get_asset_path(
    run_path: Optional[str],
    kind: Optional[str] = None,
    name: Optional[str] = None,
    step: Optional[int] = None,
    ext: Optional[str] = None,
) -> str:
    if not run_path:
        raise ValueError("run_path must be provided to get asset path.")
    _path = get_event_assets_path(run_path, kind)
    if name:
        _path = "{}/{}".format(_path, name)
    if step is not None:
        _path = "{}_{}".format(_path, step)
    if ext:
        _path = "{}.{}".format(_path, ext)

    return _path
